/*
 * This file is part of Direction, which is an AI Library for OpenTTD
 * Copyright (C) 2010  Leif Linse
 *
 * Direction is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * Direction is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Direction; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

class Direction
{
	//////////////////////////////////////////////////////////////////////
	//                                                                  //
	//  Constants                                                       //
	//                                                                  //
	//////////////////////////////////////////////////////////////////////

	/* Direction constants */
	static DIR_N = 0;
	static DIR_NE = 1;
	static DIR_E  = 2;
	static DIR_SE = 3;
	static DIR_S  = 4;
	static DIR_SW = 5;
	static DIR_W  = 6;
	static DIR_NW = 7;

	/* Special values */
	static DIR_FIRST = 0;
	static DIR_LAST = 7;

	static DIR_INVALID = 8;

	//////////////////////////////////////////////////////////////////////
	//                                                                  //
	//  To string                                                       //
	//                                                                  //
	//////////////////////////////////////////////////////////////////////

	/* Translates a direction value into a human readable string that 
	 * can be used for logging.
	 */
	static function GetDirString(dir);


	//////////////////////////////////////////////////////////////////////
	//                                                                  //
	//  Classification of direction                                     //
	//                                                                  //
	//////////////////////////////////////////////////////////////////////

	/* Returns true if the direction is a main direction (NW, SW, SE or NE) */
	static function IsMainDir(dir);

	/* Returns true if the direction is a diagonal direction (N, W, S or E) */
	static function IsDiagonalDir(dir);


	//////////////////////////////////////////////////////////////////////
	//                                                                  //
	//  Move In Direction                                               //
	//                                                                  //
	//////////////////////////////////////////////////////////////////////

	/* Returns the tile you get to if you move one tile in the 
	 * 'in_direction' direction from 'from_tile'.
	 */ 
	static function GetAdjacentTileInDirection(from_tile, in_direction);


	//////////////////////////////////////////////////////////////////////
	//                                                                  //
	//  Get Direction between tiles                                     //
	//                                                                  //
	//////////////////////////////////////////////////////////////////////

	/* If these function fails they return Direction.DIR_INVALID */

	/* Returns the direction you have to go from tile in order to reach
	 * adjacent_tile. A precondition is that these two tiles are
	 * adjacent.
	 */
	static function GetDirectionToAdjacentTile(tile, adjacent_tile);

	/* Returns the direction you have to move from tile1 to reach tile2.
	 * This function do not require tile1 and tile2 to be adjacent, but 
	 * they must be exactly in one of the eight directions.
	 */
	static function GetDirectionToTile(tile1, tile2);


	//////////////////////////////////////////////////////////////////////
	//                                                                  //
	//  Turn Direction                                                  //
	//                                                                  //
	//////////////////////////////////////////////////////////////////////
	
	/* Turns 'dir' 45 degrees clockwise a given number of times. 
	 * num_45_deg must be an integer value. Negative values of num_45_deg
	 * are accepted.
	 */
	static function TurnDirClockwise45Deg(dir, num_45_deg);

	/* Same as TurnDirClockwise45Deg but turns anti clocwise instead */
	static function TurnDirAntiClockwise45Deg(dir, num_45_deg);


	//////////////////////////////////////////////////////////////////////

	/* Only intended for use within the Direction library.
	 * If you want to use the util.tile class in your AI,
	 * then make your own import. */
	//static _Tile = import("util.tile", "", 1);

	/* Because libraries can't be cyclic dependant a copy of the GetTileRelative 
	 * function is needed in this class. Please use Tile.GetTileRelative(..) in
	 * your AI code.
	 */
	static function GetTileRelative(relative_to_tile, delta_x, delta_y);
}

function Direction::GetDirString(dir)
{
	switch(dir)
	{
		case Direction.DIR_N:
			return "N";
		case Direction.DIR_NE:
			return "NE";
		case Direction.DIR_E:
			return "E";
		case Direction.DIR_SE:
			return "SE";
		case Direction.DIR_S:
			return "S";
		case Direction.DIR_SW:
			return "SW";
		case Direction.DIR_W:
			return "W";
		case Direction.DIR_NW:
			return "NW";
		default:
			AILog.Error("Direction::GetDirString: invalid direction: " + in_direction);
			return -1;
	}
	
}

function Direction::IsMainDir(dir)
{
	return dir == Direction.DIR_NE ||
			dir == Direction.DIR_SE ||
			dir == Direction.DIR_SW ||
			dir == Direction.DIR_NW;
}

function Direction::IsDiagonalDir(dir)
{
	return dir == Direction.DIR_N ||
			dir == Direction.DIR_E ||
			dir == Direction.DIR_S ||
			dir == Direction.DIR_W;
}

function Direction::GetAdjacentTileInDirection(from_tile, in_direction)
{
	switch(in_direction)
	{
		case Direction.DIR_N:
			return Direction.GetTileRelative(from_tile, -1, -1);
		case Direction.DIR_NE:
			return Direction.GetTileRelative(from_tile, -1,  0);
		case Direction.DIR_E:
			return Direction.GetTileRelative(from_tile, -1,  1);
		case Direction.DIR_SE:
			return Direction.GetTileRelative(from_tile,  0,  1);
		case Direction.DIR_S:
			return Direction.GetTileRelative(from_tile,  1,  1);
		case Direction.DIR_SW:
			return Direction.GetTileRelative(from_tile,  1,  0);
		case Direction.DIR_W:
			return Direction.GetTileRelative(from_tile,  1, -1);
		case Direction.DIR_NW:
			return Direction.GetTileRelative(from_tile,  0, -1);
		default:
			AILog.Error("Direction::GetAdjacentTileInDirection: invalid direction: " + in_direction);
			return -1;
	}
}

function Direction::GetDirectionToAdjacentTile(tile, adjacent_tile)
{
	local rel_x = AIMap.GetTileX(adjacent_tile) - AIMap.GetTileX(tile);
	local rel_y = AIMap.GetTileY(adjacent_tile) - AIMap.GetTileY(tile);

	local rel_dir = [ [Direction.DIR_N, [-1, -1]],
			[Direction.DIR_NE, [-1,  0]],
			[Direction.DIR_E,  [-1,  1]],
			[Direction.DIR_SE, [ 0,  1]],
			[Direction.DIR_S,  [ 1,  1]],
			[Direction.DIR_SW, [ 1,  0]],
			[Direction.DIR_W,  [ 1, -1]],
			[Direction.DIR_NW, [ 0, -1]] ];

	foreach(dir in rel_dir)
	{
		local rel = dir[1];
		local X = 0;
		local Y = 1;

		if(rel[X] == rel_x && rel[Y] == rel_y)
		{
			return dir[0];
		}
	}

	return Direction.DIR_INVALID;
}

function Direction::GetDirectionToTile(tile1, tile2)
{
	local rel_x = AIMap.GetTileX(tile1) - AIMap.GetTileX(tile2);
	local rel_y = AIMap.GetTileY(tile1) - AIMap.GetTileY(tile2);

	if(rel_x == 0 && rel_y == 0)
	{
		return Direction.DIR_INVALID;
	}
	else if (Helper.Abs(rel_x) >= 1 && Helper.Abs(rel_y) >= 1)
	{
		// Neither of NE, NW, SW, SE

		if (Helper.Abs(rel_x) == Helper.Abs(rel_y))
		{
			// Same amplitude of rel_x and rel_y => N, W, S or E.
			if(rel_x <= -1 && rel_y <= -1) return Direction.DIR_N;
			if(rel_x <= -1 && rel_y >= 1) return Direction.DIR_E;
			if(rel_x >= 1 && rel_y >= 1) return Direction.DIR_S;
			if(rel_x >= 1 && rel_y <= -1) return Direction.DIR_S;

			// Error
			return Direction.DIR_INVALID;
		}
		else
		{
			// Not a valid direction
			return Direction.DIR_INVALID;
		}
	}
	else
	{
		// either rel_x or rel_y is 0 and the other has a value > 0.
		
		if (rel_x < 0) return Direction.DIR_NE;
		if (rel_y > 0) return Direction.DIR_SE;
		if (rel_x > 0) return Direction.DIR_SW;
		if (rel_y < 0) return Direction.DIR_NW;

		// Error
		return Direction.DIR_INVALID
	}
		
	// Error
	return Direction.DIR_INVALID
}

function Direction::TurnDirClockwise45Deg(dir, num_45_deg)
{
	/* Make sure to only handle positive turns */
	if(num_45_deg < 0) return TurnDirAntiClockwise45Deg(dir, -num_45_deg);

	dir += num_45_deg
	
	while(dir > Direction.DIR_LAST)
	{
		dir -= (Direction.DIR_LAST - Direction.DIR_FIRST) + 1;
	} 

	return dir;
}

function Direction::TurnDirAntiClockwise45Deg(dir, num_45_deg)
{
	/* Make sure to only handle positive turns */
	if(num_45_deg < 0) return TurnDirClockwise45Deg(dir, -num_45_deg);

	dir -= num_45_deg
	
	while(dir < Direction.DIR_FIRST)
	{
		dir += (Direction.DIR_LAST - Direction.DIR_FIRST) + 1;
	}

	return dir;
}

function Direction::GetTileRelative(relative_to_tile, delta_x, delta_y)
{
	local tile_x = AIMap.GetTileX(relative_to_tile);
	local tile_y = AIMap.GetTileY(relative_to_tile);

	local new_x = Tile._Helper.Clamp(tile_x + delta_x, 1, AIMap.GetMapSizeX() - 2);
	local new_y = Tile._Helper.Clamp(tile_y + delta_y, 1, AIMap.GetMapSizeY() - 2);

	return AIMap.GetTileIndex(new_x, new_y);
}
